# Workaround for conflict with distributed 1.23.0
# https://github.com/dask/dask-xgboost/pull/27#issuecomment-417474734
from concurrent.futures import ThreadPoolExecutor

import dask
import dask.array as da
import dask.dataframe as dd
import distributed.comm.utils
import numpy as np
import pandas as pd
import pytest
import scipy.sparse
import xgboost as xgb
from dask.array.utils import assert_eq
from dask.distributed import Client
from distributed.utils_test import cluster, gen_cluster, loop  # noqa
from sklearn.datasets import load_digits, load_iris
from sklearn.model_selection import train_test_split

import dask_xgboost as dxgb
from dask_xgboost.core import _package_evals

distributed.comm.utils._offload_executor = ThreadPoolExecutor(max_workers=2)


df = pd.DataFrame(
    {"x": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "y": [1, 0, 1, 0, 1, 0, 1, 0, 1, 0],}
)
labels = pd.Series([1, 0, 1, 0, 1, 0, 1, 1, 1, 1])

param = {
    "max_depth": 2,
    "eta": 1,
    "silent": 1,
    "objective": "binary:logistic",
}

X = df.values
y = labels.values


def test_classifier(loop):  # noqa
    digits = load_digits(2)
    X = digits["data"]
    y = digits["target"]

    with cluster() as (s, [a, b]):
        with Client(s["address"], loop=loop):
            a = dxgb.XGBClassifier()
            X2 = da.from_array(X)
            y2 = da.from_array(y)
            a.fit(X2, y2)
            p1 = a.predict(X2)

    b = xgb.XGBClassifier()
    b.fit(X, y)
    np.testing.assert_array_almost_equal(a.feature_importances_, b.feature_importances_)
    assert_eq(p1, b.predict(X))


def test_classifier_different_chunks(loop):  # noqa
    with cluster() as (s, [a, b]):
        with Client(s["address"], loop=loop):
            a = dxgb.XGBClassifier()
            X2 = da.from_array(X, 5)
            y2 = da.from_array(y, 4)

            with pytest.raises(ValueError):
                a.fit(X2, y2)


def test_multiclass_classifier(loop):  # noqa
    # data
    iris = load_iris()
    X, y = iris.data, iris.target
    dX = da.from_array(X, 5)
    dy = da.from_array(y, 5)
    df = pd.DataFrame(X, columns=iris.feature_names)
    labels = pd.Series(y, name="target")

    ddf = dd.from_pandas(df, 2)
    dlabels = dd.from_pandas(labels, 2)
    # model
    a = xgb.XGBClassifier()  # array
    b = dxgb.XGBClassifier()
    c = xgb.XGBClassifier()  # frame
    d = dxgb.XGBClassifier()

    with cluster() as (s, [_, _]):
        with Client(s["address"], loop=loop):
            # fit
            a.fit(X, y)  # array
            b.fit(dX, dy, classes=[0, 1, 2])
            c.fit(df, labels)  # frame
            d.fit(ddf, dlabels, classes=[0, 1, 2])

            # check
            da.utils.assert_eq(a.predict(X), b.predict(dX))
            da.utils.assert_eq(a.predict_proba(X), b.predict_proba(dX))
            da.utils.assert_eq(c.predict(df), d.predict(ddf))
            da.utils.assert_eq(c.predict_proba(df), d.predict_proba(ddf))


def test_classifier_early_stopping(loop):  # noqa
    # data
    digits = load_digits(2)
    X = digits["data"]
    y = digits["target"]
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

    dX_train = da.from_array(X_train)
    dy_train = da.from_array(y_train)

    clf1 = dxgb.XGBClassifier()
    clf2 = dxgb.XGBClassifier()
    clf3 = dxgb.XGBClassifier()
    with cluster() as (s, [_, _]):
        with Client(s["address"], loop=loop):
            clf1.fit(
                dX_train,
                dy_train,
                early_stopping_rounds=5,
                eval_metric="auc",
                eval_set=[(X_test, y_test)],
            )
            clf2.fit(
                dX_train,
                dy_train,
                early_stopping_rounds=4,
                eval_metric="auc",
                eval_set=[(X_test, y_test)],
            )

            # should be the same
            assert clf1.best_score == clf2.best_score
            assert clf1.best_score != 1

            # check overfit
            clf3.fit(
                dX_train,
                dy_train,
                early_stopping_rounds=10,
                eval_metric="auc",
                eval_set=[(X_test, y_test)],
            )
            assert clf3.best_score == 1


def test_package_evals():
    # data
    digits = load_digits(2)
    X = digits["data"]
    y = digits["target"]
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

    evals = _package_evals([(X_test, y_test), (X, y_test)])

    assert len(evals) == 2

    evals = _package_evals(
        [(X_test, y_test), (X, y_test)], sample_weight_eval_set=[[1], [2]]
    )

    assert len(evals) == 2

    evals = _package_evals(
        [(X_test, y_test), (X, y_test)], sample_weight_eval_set=[[1]]
    )

    assert len(evals) == 1


def test_validation_weights_xgbclassifier(loop):  # noqa
    from sklearn.datasets import make_hastie_10_2

    # prepare training and test data
    X, y = make_hastie_10_2(n_samples=2000, random_state=42)
    labels, y = np.unique(y, return_inverse=True)

    param_dist = {
        "objective": "binary:logistic",
        "n_estimators": 2,
        "random_state": 123,
    }

    with cluster() as (s, [a, b]):
        with Client(s["address"], loop=loop):
            X_train, X_test = X[:1600], X[1600:]
            y_train, y_test = y[:1600], y[1600:]

            dX_train = da.from_array(X_train)
            dy_train = da.from_array(y_train)

            # instantiate model
            clf = dxgb.XGBClassifier(**param_dist)

            # train it using instance weights only in the training set
            weights_train = np.random.choice([1, 2], len(X_train))
            weights_train = da.from_array(weights_train)
            clf.fit(
                dX_train,
                dy_train,
                sample_weight=weights_train,
                eval_set=[(X_test, y_test)],
                eval_metric="logloss",
            )

            # evaluate logloss metric on test set *without* using weights
            evals_result_without_weights = clf.evals_result()
            logloss_without_weights = evals_result_without_weights["validation_0"][
                "logloss"
            ]

            # now use weights for the test set
            np.random.seed(0)
            weights_test = np.random.choice([1, 2], len(X_test))
            clf.fit(
                dX_train,
                dy_train,
                sample_weight=weights_train,
                eval_set=[(X_test, y_test)],
                sample_weight_eval_set=[weights_test],
                eval_metric="logloss",
            )
            evals_result_with_weights = clf.evals_result()
            logloss_with_weights = evals_result_with_weights["validation_0"]["logloss"]

    # check that the logloss in the test set is actually different
    # when using weights than when not using them
    assert all((logloss_with_weights[i] != logloss_without_weights[i] for i in [0, 1]))


@pytest.mark.parametrize("kind", ["array", "dataframe"])
def test_classifier_multi(kind, loop):  # noqa: F811

    if kind == "array":
        X2 = da.from_array(X, 5)
        y2 = da.from_array(np.array([0, 1, 2, 0, 1, 2, 0, 0, 0, 1]), chunks=5)
    else:
        X2 = dd.from_pandas(df, npartitions=2)
        y2 = dd.from_pandas(labels, npartitions=2)

    with cluster() as (s, [a, b]):
        with Client(s["address"], loop=loop):
            a = dxgb.XGBClassifier(
                num_class=3, n_estimators=10, objective="multi:softprob"
            )
            a.fit(X2, y2)
            p1 = a.predict(X2)

            assert dask.is_dask_collection(p1)

            if kind == "array":
                assert p1.shape == (10,)

            result = p1.compute()
            assert result.shape == (10,)

            # proba
            p2 = a.predict_proba(X2)
            assert dask.is_dask_collection(p2)

            if kind == "array":
                assert p2.shape == (10, 3)
            assert p2.compute().shape == (10, 3)


def test_regressor(loop):  # noqa
    with cluster() as (s, [a, b]):
        with Client(s["address"], loop=loop):
            a = dxgb.XGBRegressor()
            X2 = da.from_array(X, 5)
            y2 = da.from_array(y, 5)
            a.fit(X2, y2)
            p1 = a.predict(X2)

    b = xgb.XGBRegressor()
    b.fit(X, y)
    assert_eq(p1, b.predict(X))


def test_regressor_with_early_stopping(loop):  # noqa
    with cluster() as (s, [a, b]):
        with Client(s["address"], loop=loop):
            a = dxgb.XGBRegressor()
            X2 = da.from_array(X, 5)
            y2 = da.from_array(y, 5)
            a.fit(
                X2, y2, early_stopping_rounds=4, eval_metric="rmse", eval_set=[(X, y)],
            )
            p1 = a.predict(X2)

    b = xgb.XGBRegressor()
    b.fit(X, y, early_stopping_rounds=4, eval_metric="rmse", eval_set=[(X, y)])
    assert_eq(p1, b.predict(X))
    assert_eq(a.best_score, b.best_score)


def test_validation_weights_xgbregressor(loop):  # noqa
    from sklearn.datasets import make_regression
    from sklearn.metrics import mean_squared_error

    # prepare training and test data
    X, y = make_regression(n_samples=2000, random_state=42)

    with cluster() as (s, [a, b]):
        with Client(s["address"], loop=loop):
            X_train, X_test = X[:1600], X[1600:]
            y_train, y_test = y[:1600], y[1600:]

            dX_train = da.from_array(X_train)
            dy_train = da.from_array(y_train)
            dX_test = da.from_array(X_test)

            reg = dxgb.XGBRegressor()

            reg.fit(
                dX_train, dy_train,  # sample_weight=weights_train,
            )
            preds = reg.predict(dX_test)

            rng = np.random.RandomState(0)
            weights_train = 100.0 + rng.rand(len(X_train))
            weights_train = da.from_array(weights_train)
            weights_test = 100.0 + rng.rand(len(X_test))

            reg.fit(
                dX_train,
                dy_train,
                sample_weight=weights_train,
                sample_weight_eval_set=[weights_test],
            )
            preds2 = reg.predict(dX_test)

    err = mean_squared_error(preds, y_test)
    err2 = mean_squared_error(preds2, y_test)
    assert err != err2


@gen_cluster(client=True, timeout=None)
def test_basic(c, s, a, b):
    dtrain = xgb.DMatrix(df, label=labels)
    bst = xgb.train(param, dtrain)

    ddf = dd.from_pandas(df, npartitions=4)
    dlabels = dd.from_pandas(labels, npartitions=4)
    dbst = yield dxgb.train(c, param, ddf, dlabels)
    dbst = yield dxgb.train(c, param, ddf, dlabels)  # we can do this twice

    result = bst.predict(dtrain)
    dresult = dbst.predict(dtrain)

    correct = (result > 0.5) == labels
    dcorrect = (dresult > 0.5) == labels
    assert dcorrect.sum() >= correct.sum()

    predictions = dxgb.predict(c, dbst, ddf)
    assert isinstance(predictions, da.Array)
    predictions = yield c.compute(predictions)._result()
    assert isinstance(predictions, np.ndarray)

    assert ((predictions > 0.5) != labels).sum() < 2


@gen_cluster(client=True, timeout=None)
def test_dmatrix_kwargs(c, s, a, b):
    xgb.rabit.init()  # workaround for "Doing rabit call after Finalize"
    dX = da.from_array(X, chunks=(2, 2))
    dy = da.from_array(y, chunks=(2,))
    dbst = yield dxgb.train(c, param, dX, dy, dmatrix_kwargs={"missing": 0.0})

    # Distributed model matches local model with dmatrix kwargs
    dtrain = xgb.DMatrix(X, label=y, missing=0.0)
    bst = xgb.train(param, dtrain)
    result = bst.predict(dtrain)
    dresult = dbst.predict(dtrain)
    assert np.abs(result - dresult).sum() < 0.02

    # Distributed model gives bad predictions without dmatrix kwargs
    dtrain_incompat = xgb.DMatrix(X, label=y)
    dresult_incompat = dbst.predict(dtrain_incompat)
    assert np.abs(result - dresult_incompat).sum() > 0.02


def _test_container(dbst, predictions, X_type):
    dtrain = xgb.DMatrix(X_type(X), label=y)
    bst = xgb.train(param, dtrain)

    result = bst.predict(dtrain)
    dresult = dbst.predict(dtrain)

    correct = (result > 0.5) == y
    dcorrect = (dresult > 0.5) == y

    assert dcorrect.sum() >= correct.sum()
    assert isinstance(predictions, np.ndarray)
    assert ((predictions > 0.5) != labels).sum() < 2


@gen_cluster(client=True, timeout=None)
def test_numpy(c, s, a, b):
    xgb.rabit.init()  # workaround for "Doing rabit call after Finalize"
    dX = da.from_array(X, chunks=(2, 2))
    dy = da.from_array(y, chunks=(2,))
    dbst = yield dxgb.train(c, param, dX, dy)
    dbst = yield dxgb.train(c, param, dX, dy)  # we can do this twice

    predictions = dxgb.predict(c, dbst, dX)
    assert isinstance(predictions, da.Array)
    predictions = yield c.compute(predictions)
    _test_container(dbst, predictions, np.array)


@gen_cluster(client=True, timeout=None)
def test_scipy_sparse(c, s, a, b):
    xgb.rabit.init()  # workaround for "Doing rabit call after Finalize"
    dX = da.from_array(X, chunks=(2, 2)).map_blocks(scipy.sparse.csr_matrix)
    dy = da.from_array(y, chunks=(2,))
    dbst = yield dxgb.train(c, param, dX, dy)
    dbst = yield dxgb.train(c, param, dX, dy)  # we can do this twice

    predictions = dxgb.predict(c, dbst, dX)
    assert isinstance(predictions, da.Array)

    predictions_result = yield c.compute(predictions)
    _test_container(dbst, predictions_result, scipy.sparse.csr_matrix)


@gen_cluster(client=True, timeout=None)
def test_sparse(c, s, a, b):
    xgb.rabit.init()  # workaround for "Doing rabit call after Finalize"
    dX = da.from_array(X, chunks=(2, 2)).map_blocks(scipy.sparse.csr_matrix)
    dy = da.from_array(y, chunks=(2,))
    dbst = yield dxgb.train(c, param, dX, dy)
    dbst = yield dxgb.train(c, param, dX, dy)  # we can do this twice

    predictions = dxgb.predict(c, dbst, dX)
    assert isinstance(predictions, da.Array)

    predictions_result = yield c.compute(predictions)
    _test_container(dbst, predictions_result, scipy.sparse.csr_matrix)


def test_synchronous_api(loop):  # noqa
    dtrain = xgb.DMatrix(df, label=labels)
    bst = xgb.train(param, dtrain)

    ddf = dd.from_pandas(df, npartitions=4)
    dlabels = dd.from_pandas(labels, npartitions=4)

    with cluster() as (s, [a, b]):
        with Client(s["address"], loop=loop) as c:

            dbst = dxgb.train(c, param, ddf, dlabels)

            result = bst.predict(dtrain)
            dresult = dbst.predict(dtrain)

            correct = (result > 0.5) == labels
            dcorrect = (dresult > 0.5) == labels
            assert dcorrect.sum() >= correct.sum()


@gen_cluster(client=True, timeout=None)
def test_errors(c, s, a, b):
    def f(part):
        raise Exception("foo")

    df = dd.demo.make_timeseries()
    df = df.map_partitions(f, meta=df._meta)

    with pytest.raises(Exception) as info:
        yield dxgb.train(c, param, df, df.x)

    assert "foo" in str(info.value)


@gen_cluster(client=True, timeout=None)
@pytest.mark.asyncio
async def test_predict_proba(c, s, a, b):
    X = da.random.random((50, 2), chunks=25)
    y = da.random.randint(0, 2, size=50, chunks=25)
    X_ = await c.compute(X)

    # array
    clf = dxgb.XGBClassifier()
    clf.fit(X, y, classes=[0, 1])
    booster = await clf._Booster

    result = clf.predict_proba(X_)
    expected = booster.predict(xgb.DMatrix(X_))
    np.testing.assert_array_equal(result, expected)

    # dataframe
    XX = dd.from_dask_array(X, columns=["A", "B"])
    yy = dd.from_dask_array(y)
    XX_ = await c.compute(XX)

    clf = dxgb.XGBClassifier()
    clf.fit(XX, yy, classes=[0, 1])
    booster = await clf._Booster

    result = clf.predict_proba(XX_)
    expected = booster.predict(xgb.DMatrix(XX_))
    np.testing.assert_array_equal(result, expected)


def test_regressor_evals_result(loop):  # noqa
    with cluster() as (s, [a, b]):
        with Client(s["address"], loop=loop):
            a = dxgb.XGBRegressor()
            X2 = da.from_array(X, 5)
            y2 = da.from_array(y, 5)
            a.fit(X2, y2, eval_metric="rmse", eval_set=[(X, y)])
            evals_result = a.evals_result()

    b = xgb.XGBRegressor()
    b.fit(X, y, eval_metric="rmse", eval_set=[(X, y)])
    assert_eq(evals_result, b.evals_result())


def test_classifier_evals_result(loop):  # noqa
    with cluster() as (s, [a, b]):
        with Client(s["address"], loop=loop):
            a = dxgb.XGBClassifier()
            X2 = da.from_array(X, 5)
            y2 = da.from_array(y, 5)
            a.fit(X2, y2, eval_metric="rmse", eval_set=[(X, y)])
            evals_result = a.evals_result()

    b = xgb.XGBClassifier()
    b.fit(X, y, eval_metric="rmse", eval_set=[(X, y)])
    assert_eq(evals_result, b.evals_result())


@gen_cluster(client=True, timeout=None)
def test_eval_set_dask_collection_exception(c, s, a, b):
    ddf = dd.from_pandas(df, npartitions=4)
    dlabels = dd.from_pandas(labels, npartitions=4)

    X2 = da.from_array(X, 5)
    y2 = da.from_array(y, 5)

    with pytest.raises(TypeError) as info:
        yield dxgb.train(c, param, ddf, dlabels, eval_set=[(X2, y2)])

    assert "Evaluation set must not contain dask collections." in str(info.value)


@gen_cluster(client=True, timeout=None)
def test_sample_weight_eval_set_dask_collection_exception(c, s, a, b):
    ddf = dd.from_pandas(df, npartitions=4)
    dlabels = dd.from_pandas(labels, npartitions=4)

    X2 = da.from_array(X, 5)
    y2 = da.from_array(y, 5)

    with pytest.raises(TypeError) as info:
        yield dxgb.train(c, param, ddf, dlabels, sample_weight_eval_set=[(X2, y2)])

    assert "Sample weight evaluation set must not contain dask collections." in str(
        info.value
    )



import logging
from collections import defaultdict
from threading import Thread

import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd
import xgboost as xgb
from dask import delayed, is_dask_collection
from dask.distributed import default_client, wait
from toolz import assoc, first
from tornado import gen

from .tracker import RabitTracker, get_host_ip

try:
    import sparse
except ImportError:
    sparse = False

try:
    import scipy.sparse as ss
except ImportError:
    ss = False

logger = logging.getLogger(__name__)


def parse_host_port(address):
    if "://" in address:
        address = address.rsplit("://", 1)[1]
    host, port = address.split(":")
    port = int(port)
    return host, port


def start_tracker(host, n_workers):
    """ Start Rabit tracker """
    if host is None:
        host = get_host_ip("auto")
    env = {"DMLC_NUM_WORKER": n_workers}
    rabit = RabitTracker(hostIP=host, nslave=n_workers)
    env.update(rabit.slave_envs())

    rabit.start(n_workers)
    logger.info("Starting Rabit Tracker")
    thread = Thread(target=rabit.join)
    thread.daemon = True
    thread.start()
    return env


def concat(L):
    if isinstance(L[0], np.ndarray):
        return np.concatenate(L, axis=0)
    elif isinstance(L[0], (pd.DataFrame, pd.Series)):
        return pd.concat(L, axis=0)
    elif ss and isinstance(L[0], ss.csr_matrix):
        return ss.vstack(L, format="csr")
    elif sparse and isinstance(L[0], sparse.SparseArray):
        return sparse.concatenate(L, axis=0)
    else:
        raise TypeError(
            "Data must be either numpy arrays or pandas dataframes"
            ". Got %s" % type(L[0])
        )


def train_part(
    env,
    param,
    list_of_parts,
    dmatrix_kwargs=None,
    eval_set=None,
    missing=None,
    n_jobs=None,
    sample_weight_eval_set=None,
    **kwargs
):
    """
    Run part of XGBoost distributed workload

    This starts an xgboost.rabit slave, trains on provided data, and then shuts
    down the xgboost.rabit slave

    Returns
    -------
    model if rank zero, None otherwise
    """
    data, labels, sample_weight = zip(*list_of_parts)  # Prepare data
    data = concat(data)  # Concatenate many parts into one
    labels = concat(labels)
    sample_weight = concat(sample_weight) if np.all(sample_weight) else None

    if dmatrix_kwargs is None:
        dmatrix_kwargs = {}

    dmatrix_kwargs["feature_names"] = getattr(data, "columns", None)
    dtrain = xgb.DMatrix(data, labels, weight=sample_weight, **dmatrix_kwargs)

    evals = _package_evals(
        eval_set,
        sample_weight_eval_set=sample_weight_eval_set,
        missing=missing,
        n_jobs=n_jobs,
    )

    args = [("%s=%s" % item).encode() for item in env.items()]
    xgb.rabit.init(args)
    try:
        local_history = {}
        logger.info("Starting Rabit, Rank %d", xgb.rabit.get_rank())
        bst = xgb.train(
            param, dtrain, evals=evals, evals_result=local_history, **kwargs
        )

        if xgb.rabit.get_rank() == 0:  # Only return from one worker
            result = bst
            evals_result = local_history
        else:
            result = None
            evals_result = None
    finally:
        logger.info("Finalizing Rabit, Rank %d", xgb.rabit.get_rank())
        xgb.rabit.finalize()
    return result, evals_result


def _package_evals(eval_set, sample_weight_eval_set=None, missing=None, n_jobs=None):
    if eval_set is not None:
        if sample_weight_eval_set is None:
            sample_weight_eval_set = [None] * len(eval_set)
        evals = list(
            xgb.DMatrix(
                data, label=label, missing=missing, weight=weight, nthread=n_jobs,
            )
            for ((data, label), weight) in zip(eval_set, sample_weight_eval_set)
        )
        evals = list(zip(evals, ["validation_{}".format(i) for i in range(len(evals))]))
    else:
        evals = ()
    return evals


def _has_dask_collections(list_of_collections, message):
    list_of_collections = list_of_collections or []
    if any(
        is_dask_collection(collection)
        for collections in list_of_collections
        for collection in collections
    ):
        raise TypeError(message)


@gen.coroutine
def _train(
    client,
    params,
    data,
    labels,
    dmatrix_kwargs={},
    evals_result=None,
    sample_weight=None,
    **kwargs
):
    """
    Asynchronous version of train

    See Also
    --------
    train
    """
    # Break apart Dask.array/dataframe into chunks/parts
    data_parts = data.to_delayed()
    label_parts = labels.to_delayed()
    if isinstance(data_parts, np.ndarray):
        assert data_parts.shape[1] == 1
        data_parts = data_parts.flatten().tolist()
    if isinstance(label_parts, np.ndarray):
        assert label_parts.ndim == 1 or label_parts.shape[1] == 1
        label_parts = label_parts.flatten().tolist()
    if sample_weight is not None:
        sample_weight_parts = sample_weight.to_delayed()
        if isinstance(sample_weight_parts, np.ndarray):
            assert sample_weight_parts.ndim == 1 or sample_weight_parts.shape[1] == 1
            sample_weight_parts = sample_weight_parts.flatten().tolist()
    else:
        # If sample_weight is None construct a list of Nones to keep
        # the structure of parts consistent.
        sample_weight_parts = [None] * len(data_parts)

    # Check that data, labels, and sample_weights are the same length
    lists = [data_parts, label_parts, sample_weight_parts]
    if len(set([len(l) for l in lists])) > 1:
        raise ValueError(
            "data, label, and sample_weight parts/chunks must have same length."
        )

    # Arrange parts into triads.  This enforces co-locality
    parts = list(map(delayed, zip(data_parts, label_parts, sample_weight_parts)))
    parts = client.compute(parts)  # Start computation in the background
    yield wait(parts)

    for part in parts:
        if part.status == "error":
            yield part  # trigger error locally

    _has_dask_collections(
        kwargs.get("eval_set", []), "Evaluation set must not contain dask collections."
    )
    _has_dask_collections(
        kwargs.get("sample_weight_eval_set", []),
        "Sample weight evaluation set must not contain dask collections.",
    )

    # Because XGBoost-python doesn't yet allow iterative training, we need to
    # find the locations of all chunks and map them to particular Dask workers
    key_to_part_dict = dict([(part.key, part) for part in parts])
    who_has = yield client.scheduler.who_has(keys=[part.key for part in parts])
    worker_map = defaultdict(list)
    for key, workers in who_has.items():
        worker_map[first(workers)].append(key_to_part_dict[key])

    ncores = yield client.scheduler.ncores()  # Number of cores per worker

    # Start the XGBoost tracker on the Dask scheduler
    env = yield client._run_on_scheduler(start_tracker, None, len(worker_map))

    # Tell each worker to train on the chunks/parts that it has locally
    futures = [
        client.submit(
            train_part,
            env,
            assoc(params, "nthread", ncores[worker]),
            list_of_parts,
            workers=worker,
            dmatrix_kwargs=dmatrix_kwargs,
            **kwargs
        )
        for worker, list_of_parts in worker_map.items()
    ]

    # Get the results, only one will be non-None
    results = yield client._gather(futures)
    result, _evals_result = [v for v in results if v.count(None) != len(v)][0]

    if evals_result is not None:
        evals_result.update(_evals_result)

    num_class = params.get("num_class")
    if num_class:
        result.set_attr(num_class=str(num_class))
    raise gen.Return(result)


def train(
    client,
    params,
    data,
    labels,
    dmatrix_kwargs={},
    evals_result=None,
    sample_weight=None,
    **kwargs
):
    """ Train an XGBoost model on a Dask Cluster

    This starts XGBoost on all Dask workers, moves input data to those workers,
    and then calls ``xgboost.train`` on the inputs.

    Parameters
    ----------
    client: dask.distributed.Client
    params: dict
        Parameters to give to XGBoost (see xgb.Booster.train)
    data: dask array or dask.dataframe
    labels: dask.array or dask.dataframe
    dmatrix_kwargs: Keywords to give to Xgboost DMatrix
    evals_result: dict, optional
        Stores the evaluation result history of all the items in the eval_set
        by mutating evals_result in place.
    sample_weight : array_like, optional
        instance weights
    **kwargs: Keywords to give to XGBoost train

    Examples
    --------
    >>> client = Client('scheduler-address:8786')  # doctest: +SKIP
    >>> data = dd.read_csv('s3://...')  # doctest: +SKIP
    >>> labels = data['outcome']  # doctest: +SKIP
    >>> del data['outcome']  # doctest: +SKIP
    >>> train(client, params, data, labels, **normal_kwargs)  # doctest: +SKIP
    <xgboost.core.Booster object at ...>

    See Also
    --------
    predict
    """
    return client.sync(
        _train,
        client,
        params,
        data,
        labels,
        dmatrix_kwargs,
        evals_result,
        sample_weight,
        **kwargs
    )


def _predict_part(part, model=None):
    xgb.rabit.init()
    try:
        dm = xgb.DMatrix(part)
        result = model.predict(dm)
    finally:
        xgb.rabit.finalize()

    if isinstance(part, pd.DataFrame):
        if model.attr("num_class"):
            result = pd.DataFrame(result, index=part.index)
        else:
            result = pd.Series(result, index=part.index, name="predictions")
    return result


def predict(client, model, data):
    """ Distributed prediction with XGBoost

    Parameters
    ----------
    client: dask.distributed.Client
    model: xgboost.Booster
    data: dask array or dataframe

    Examples
    --------
    >>> client = Client('scheduler-address:8786')  # doctest: +SKIP
    >>> test_data = dd.read_csv('s3://...')  # doctest: +SKIP
    >>> model
    <xgboost.core.Booster object at ...>

    >>> predictions = predict(client, model, test_data)  # doctest: +SKIP

    Returns
    -------
    Dask.dataframe or dask.array, depending on the input data type

    See Also
    --------
    train
    """
    if isinstance(data, dd._Frame):
        result = data.map_partitions(_predict_part, model=model)
        result = result.values
    elif isinstance(data, da.Array):
        num_class = model.attr("num_class") or 2
        num_class = int(num_class)

        if num_class > 2:
            kwargs = dict(drop_axis=None, chunks=(data.chunks[0], (num_class,)))
        else:
            kwargs = dict(drop_axis=1)
        result = data.map_blocks(_predict_part, model=model, dtype=np.float32, **kwargs)
    else:
        model = model.result()  # Future to concrete
        if not isinstance(data, xgb.DMatrix):
            data = xgb.DMatrix(data)
        result = model.predict(data)

    return result


class XGBRegressor(xgb.XGBRegressor):
    def fit(
        self,
        X,
        y=None,
        eval_set=None,
        sample_weight=None,
        sample_weight_eval_set=None,
        eval_metric=None,
        early_stopping_rounds=None,
    ):
        """Fit the gradient boosting model

        Parameters
        ----------
        X : array-like [n_samples, n_features]
        y : array-like

        Returns
        -------
        self : the fitted Regressor

        Notes
        -----
        This differs from the XGBoost version not supporting the ``eval_set``,
        ``eval_metric``, ``early_stopping_rounds`` and ``verbose`` fit
        kwargs.
        eval_set : list, optional
            A list of (X, y) tuple pairs to use as validation sets, for which
            metrics will be computed.
            Validation metrics will help us track the performance of the model.
        sample_weight : array_like, optional
            instance weights
        sample_weight_eval_set : list, optional
            A list of the form [L_1, L_2, ..., L_n], where each L_i is a list
            of instance weights on the i-th validation set.
        eval_metric : str, list of str, or callable, optional
            If a str, should be a built-in evaluation metric to use. See
            `doc/parameter.rst <https://github.com/dmlc/xgboost/blob/master/doc/parameter.rst>`_.  # noqa: E501
            If a list of str, should be the list of multiple built-in
            evaluation metrics to use.
            If callable, a custom evaluation metric. The call
            signature is ``func(y_predicted, y_true)`` where ``y_true`` will
            be a DMatrix object such that you may need to call the
            ``get_label`` method. It must return a str, value pair where
            the str is a name for the evaluation and value is the value of
            the evaluation function. The callable custom objective is always
            minimized.
        early_stopping_rounds : int
            Activates early stopping. Validation metric needs to improve at
            least once in every **early_stopping_rounds** round(s) to continue
            training.
            Requires at least one item in **eval_set**.
            The method returns the model from the last iteration (not the best
            one).
            If there's more than one item in **eval_set**, the last entry will
            be used for early stopping.
            If there's more than one metric in **eval_metric**, the last
            metric will be used for early stopping.
            If early stopping occurs, the model will have three additional
            fields:
            ``clf.best_score``, ``clf.best_iteration`` and
            ``clf.best_ntree_limit``.
        """
        client = default_client()
        xgb_options = self.get_xgb_params()

        if eval_metric is not None:
            if callable(eval_metric):
                eval_metric = None
            else:
                xgb_options.update({"eval_metric": eval_metric})

        self.evals_result_ = {}
        self._Booster = train(
            client,
            xgb_options,
            X,
            y,
            num_boost_round=self.n_estimators,
            eval_set=eval_set,
            sample_weight=sample_weight,
            sample_weight_eval_set=sample_weight_eval_set,
            missing=self.missing,
            n_jobs=self.n_jobs,
            early_stopping_rounds=early_stopping_rounds,
            evals_result=self.evals_result_,
        )

        if early_stopping_rounds is not None:
            self.best_score = self._Booster.best_score
            self.best_iteration = self._Booster.best_iteration
            self.best_ntree_limit = self._Booster.best_ntree_limit
        return self

    def predict(self, X):
        client = default_client()
        return predict(client, self._Booster, X)


class XGBClassifier(xgb.XGBClassifier):
    def fit(
        self,
        X,
        y=None,
        classes=None,
        eval_set=None,
        sample_weight=None,
        sample_weight_eval_set=None,
        eval_metric=None,
        early_stopping_rounds=None,
    ):
        """Fit a gradient boosting classifier

        Parameters
        ----------
        X : array-like [n_samples, n_features]
            Feature Matrix. May be a dask.array or dask.dataframe
        y : array-like
            Labels
        eval_set : list, optional
            A list of (X, y) tuple pairs to use as validation sets, for which
            metrics will be computed.
            Validation metrics will help us track the performance of the model.
        sample_weight : array_like, optional
            instance weights
        sample_weight_eval_set : list, optional
            A list of the form [L_1, L_2, ..., L_n], where each L_i is a list
            of instance weights on the i-th validation set.
        eval_metric : str, list of str, or callable, optional
            If a str, should be a built-in evaluation metric to use. See
            `doc/parameter.rst <https://github.com/dmlc/xgboost/blob/master/doc/parameter.rst>`_.  # noqa: E501
            If a list of str, should be the list of multiple built-in
            evaluation metrics to use.
            If callable, a custom evaluation metric. The call
            signature is ``func(y_predicted, y_true)`` where ``y_true`` will
            be a DMatrix object such that you may need to call the
            ``get_label`` method. It must return a str, value pair where
            the str is a name for the evaluation and value is the value of
            the evaluation function. The callable custom objective is always
            minimized.
        early_stopping_rounds : int
            Activates early stopping. Validation metric needs to improve at
            least once in every **early_stopping_rounds** round(s) to continue
            training.
            Requires at least one item in **eval_set**.
            The method returns the model from the last iteration (not the best
            one).
            If there's more than one item in **eval_set**, the last entry will
            be used for early stopping.
            If there's more than one metric in **eval_metric**, the last
            metric will be used for early stopping.
            If early stopping occurs, the model will have three additional
            fields:
            ``clf.best_score``, ``clf.best_iteration`` and
            ``clf.best_ntree_limit``.
        classes : sequence, optional
            The unique values in `y`. If no specified, this will be
            eagerly computed from `y` before training.

        Returns
        -------
        self : XGBClassifier

        Notes
        -----
        This differs from the XGBoost version in three ways

        1. The ``verbose`` fit kwargs are not supported.
        2. The labels are not automatically label-encoded
        3. The ``classes_`` and ``n_classes_`` attributes are not learned
        """
        client = default_client()

        if classes is None:
            if isinstance(y, da.Array):
                classes = da.unique(y)
            else:
                classes = y.unique()
            classes = classes.compute()
        else:
            classes = np.asarray(classes)
        self.classes_ = classes
        self.n_classes_ = len(self.classes_)

        xgb_options = self.get_xgb_params()

        if eval_metric is not None:
            if callable(eval_metric):
                eval_metric = None
            else:
                xgb_options.update({"eval_metric": eval_metric})

        if self.n_classes_ > 2:
            # xgboost just ignores the user-provided objective
            # We only overwrite if it's the default...
            if xgb_options["objective"] == "binary:logistic":
                xgb_options["objective"] = "multi:softprob"

            xgb_options.setdefault("num_class", self.n_classes_)

        # xgboost sets this to self.objective, which I think is wrong
        # hyper-parameters should not be updated during fit.
        self.objective = xgb_options["objective"]

        # TODO: auto label-encode y
        # that will require a dependency on dask-ml

        self.evals_result_ = {}
        self._Booster = train(
            client,
            xgb_options,
            X,
            y,
            num_boost_round=self.n_estimators,
            eval_set=eval_set,
            sample_weight=sample_weight,
            sample_weight_eval_set=sample_weight_eval_set,
            missing=self.missing,
            n_jobs=self.n_jobs,
            early_stopping_rounds=early_stopping_rounds,
            evals_result=self.evals_result_,
        )

        if early_stopping_rounds is not None:
            self.best_score = self._Booster.best_score
            self.best_iteration = self._Booster.best_iteration
            self.best_ntree_limit = self._Booster.best_ntree_limit
        return self

    def predict(self, X):
        client = default_client()
        class_probs = predict(client, self._Booster, X)
        if class_probs.ndim > 1:
            cidx = da.argmax(class_probs, axis=1)
        else:
            cidx = (class_probs > 0.5).astype(np.int64)
        return cidx

    def predict_proba(self, data, ntree_limit=None):
        client = default_client()
        if ntree_limit is not None:
            raise NotImplementedError("'ntree_limit' is not currently " "supported.")
        class_probs = predict(client, self._Booster, data)
        return class_probs



#!/usr/bin/env python

import os
from setuptools import setup

requires = open("requirements.txt").read().strip().split("\n")
install_requires = []
extras_require = {"sparse": ["sparse", "scipy"]}
for r in requires:
    if ";" in r:
        # requirements.txt conditional dependencies need to be reformatted for
        # wheels to the form: `'[extra_name]:condition' : ['requirements']`
        req, cond = r.split(";", 1)
        cond = ":" + cond
        cond_reqs = extras_require.setdefault(cond, [])
        cond_reqs.append(req)
    else:
        install_requires.append(r)

setup(
    name="dask-xgboost",
    version="0.1.11",
    description="Interactions between Dask and XGBoost",
    maintainer="Matthew Rocklin",
    maintainer_email="mrocklin@continuum.io",
    url="https://github.com/dask/dask-xgboost",
    license="BSD",
    install_requires=install_requires,
    extras_require=extras_require,
    packages=["dask_xgboost"],
    long_description=(
        open("README.rst").read() if os.path.exists("README.rst") else ""
    ),
    zip_safe=False,
)
